@@ -1,3 +1,6 @@
+// Copyright (c) HashiCorp, Inc.
+// SPDX-License-Identifier: MPL-2.0
+
 package sourcebundle
 
 import (
@@ -17,8 +20,9 @@ import (
 	"github.com/apparentlymart/go-versions/versions"
 	"github.com/apparentlymart/go-versions/versions/constraints"
 	"github.com/google/go-cmp/cmp"
-	"github.com/hashicorp/go-slug/sourceaddrs"
 	regaddr "github.com/hashicorp/terraform-registry-address"
+
+	"github.com/hashicorp/go-slug/sourceaddrs"
 )
 
 func TestBuilderSimple(t *testing.T) {
@@ -41,6 +45,7 @@ func TestBuilderSimple(t *testing.T) {
 				"1.0.0": "https://example.com/foo.tgz",
 			},
 		},
+		nil,
 	)
 
 	realSource := sourceaddrs.MustParseSource("https://example.com/foo.tgz").(sourceaddrs.RemoteSource)
@@ -111,6 +116,7 @@ func TestBuilderSubdirs(t *testing.T) {
 				"1.0.0": "https://example.com/subdirs.tgz//a",
 			},
 		},
+		nil,
 	)
 
 	// NOTE: We're asking for subdir "b" of the registry address. That combines
@@ -180,6 +186,7 @@ func TestBuilderRemoteDeps(t *testing.T) {
 			"https://example.com/dependency2.tgz": "testdata/pkgs/terraformignore",
 		},
 		nil,
+		nil,
 	)
 
 	startSource := sourceaddrs.MustParseSource("https://example.com/with-deps.tgz").(sourceaddrs.RemoteSource)
@@ -266,6 +273,92 @@ func TestBuilderRemoteDeps(t *testing.T) {
 	})
 }
 
+func TestBuilderRegistryVersionDeprecation(t *testing.T) {
+	// This tests the common pattern of specifying a module registry address
+	// to start, having that translated into a real remote source address,
+	// and then downloading from that real source address. There are no
+	// oddities or edge-cases here.
+
+	tracer := testBuildTracer{}
+	ctx := tracer.OnContext(context.Background())
+
+	targetDir := t.TempDir()
+	builder := testingBuilder(
+		t, targetDir,
+		map[string]string{
+			"https://example.com/foo.tgz": "testdata/pkgs/hello",
+		},
+		map[string]map[string]string{
+			"example.com/foo/bar/baz": {
+				"1.0.0": "https://example.com/foo.tgz",
+			},
+		},
+		map[string]map[string]*ModulePackageVersionDeprecation{
+			"example.com/foo/bar/baz": {
+				"1.0.0": &ModulePackageVersionDeprecation{
+					Reason: "test reason",
+					Link:   "test link",
+				},
+			},
+		},
+	)
+
+	realSource := sourceaddrs.MustParseSource("https://example.com/foo.tgz").(sourceaddrs.RemoteSource)
+	regSource := sourceaddrs.MustParseSource("example.com/foo/bar/baz").(sourceaddrs.RegistrySource)
+	diags := builder.AddRegistrySource(ctx, regSource, versions.All, noDependencyFinder)
+	if len(diags) > 0 {
+		t.Fatal("unexpected diagnostics")
+	}
+
+	bundle, err := builder.Close()
+
+	if err != nil {
+		t.Fatalf("failed to close bundle: %s", err)
+	}
+
+	version, _ := versions.ParseVersion("1.0.0")
+	pkgAddr, _ := sourceaddrs.ParseRegistryPackage("example.com/foo/bar/baz")
+
+	wantDeprecations := map[regaddr.ModulePackage]map[versions.Version]*RegistryVersionDeprecation{
+		pkgAddr: {
+			version: &RegistryVersionDeprecation{
+				Version: "1.0.0",
+				Reason:  "test reason",
+				Link:    "test link",
+			},
+		},
+	}
+	gotDeprecations := bundle.registryPackageVersionDeprecations
+	if diff := cmp.Diff(wantDeprecations, gotDeprecations); diff != "" {
+		t.Errorf("wrong deprecations\n%s", diff)
+	}
+
+	localPkgDir, err := bundle.LocalPathForRemoteSource(realSource)
+	if err != nil {
+		for pkgAddr, localDir := range builder.remotePackageDirs {
+			t.Logf("contents of %s are in %s", pkgAddr, localDir)
+		}
+		t.Fatalf("builder does not know a local directory for %s: %s", realSource.Package(), err)
+	}
+
+	if info, err := os.Lstat(filepath.Join(localPkgDir, "hello")); err != nil {
+		t.Errorf("problem with output file: %s", err)
+	} else if !info.Mode().IsRegular() {
+		t.Errorf("output file is not a regular file")
+	}
+
+	// Looking up the original registry address at the selected version
+	// should return the same directory, because the registry address is just
+	// an indirection over the same source address.
+	registryPkgDir, err := bundle.LocalPathForRegistrySource(regSource, versions.MustParseVersion("1.0.0"))
+	if err != nil {
+		t.Fatalf("builder does not know a local directory for %s: %s", regSource.Package(), err)
+	}
+	if registryPkgDir != localPkgDir {
+		t.Errorf("local dir for %s doesn't match local dir for %s", regSource, realSource)
+	}
+}
+
 func TestBuilderRemoteDepsDifferingTypes(t *testing.T) {
 	tracer := testBuildTracer{}
 	ctx := tracer.OnContext(context.Background())
@@ -279,6 +372,7 @@ func TestBuilderRemoteDepsDifferingTypes(t *testing.T) {
 			"https://example.com/dependency2.tgz":     "testdata/pkgs/terraformignore",
 		},
 		nil,
+		nil,
 	)
 
 	startSource := sourceaddrs.MustParseSource("https://example.com/self_dependency.tgz").(sourceaddrs.RemoteSource)
@@ -380,6 +474,7 @@ func TestBuilderTerraformIgnore(t *testing.T) {
 			"https://example.com/ignore.tgz": "testdata/pkgs/terraformignore",
 		},
 		nil,
+		nil,
 	)
 
 	startSource := sourceaddrs.MustParseSource("https://example.com/ignore.tgz").(sourceaddrs.RemoteSource)
@@ -439,6 +534,7 @@ func TestBuilderCoalescePackages(t *testing.T) {
 			"https://example.com/dependency2.tgz": "testdata/pkgs/hello",
 		},
 		nil,
+		nil,
 	)
 
 	startSource := sourceaddrs.MustParseSource("https://example.com/with-deps.tgz").(sourceaddrs.RemoteSource)
@@ -539,7 +635,7 @@ func TestBuilderCoalescePackages(t *testing.T) {
 	})
 }
 
-func testingBuilder(t *testing.T, targetDir string, remotePackages map[string]string, registryPackages map[string]map[string]string) *Builder {
+func testingBuilder(t *testing.T, targetDir string, remotePackages map[string]string, registryPackages map[string]map[string]string, registryVersionDeprecations map[string]map[string]*ModulePackageVersionDeprecation) *Builder {
 	t.Helper()
 
 	type fakeRemotePackage struct {
@@ -554,6 +650,7 @@ func testingBuilder(t *testing.T, targetDir string, remotePackages map[string]st
 
 	remotePkgs := make([]fakeRemotePackage, 0, len(remotePackages))
 	registryPkgs := make([]fakeRegistryPackage, 0, len(registryPackages))
+	registryDeprecations := make(map[string]map[versions.Version]*ModulePackageVersionDeprecation)
 
 	for pkgAddrRaw, localDir := range remotePackages {
 		pkgAddr, err := sourceaddrs.ParseRemotePackage(pkgAddrRaw)
@@ -590,7 +687,25 @@ func testingBuilder(t *testing.T, targetDir string, remotePackages map[string]st
 		registryPkgs = append(registryPkgs, pkg)
 	}
 
-	fetcher := packageFetcherFunc(func(ctx context.Context, sourceType string, url *url.URL, targetDir string) (*PackageMeta, error) {
+	for pkgAddrRaw, deprecations := range registryVersionDeprecations {
+		pkgAddr, err := sourceaddrs.ParseRegistryPackage(pkgAddrRaw)
+		if err != nil {
+			t.Fatalf("invalid registry package address %q: %s", pkgAddrRaw, err)
+		}
+
+		for versionRaw, versionDeprecation := range deprecations {
+			version, err := versions.ParseVersion(versionRaw)
+			if err != nil {
+				t.Fatalf("invalid registry package version %q for %s: %s", versionRaw, pkgAddr, err)
+			}
+			registryDeprecations[pkgAddr.Namespace] = map[versions.Version]*ModulePackageVersionDeprecation{
+				version: versionDeprecation,
+			}
+		}
+	}
+
+	fetcher := packageFetcherFunc(func(ctx context.Context, sourceType string, url *url.URL, targetDir string) (FetchSourcePackageResponse, error) {
+		var ret FetchSourcePackageResponse
 		// Our fake implementation of "fetching" is to just copy one local
 		// directory into another.
 		for _, pkg := range remotePkgs {
@@ -603,39 +718,45 @@ func testingBuilder(t *testing.T, targetDir string, remotePackages map[string]st
 			localDir := pkg.localDir
 			err := copyDir(targetDir, localDir)
 			if err != nil {
-				return nil, fmt.Errorf("copying %s to %s: %w", localDir, targetDir, err)
+				return ret, fmt.Errorf("copying %s to %s: %w", localDir, targetDir, err)
 			}
-			return nil, nil
+			return ret, nil
 		}
-		return nil, fmt.Errorf("no fake remote package matches %s %s", sourceType, url)
+		return ret, fmt.Errorf("no fake remote package matches %s %s", sourceType, url)
 	})
 
 	registryClient := registryClientFuncs{
-		modulePackageVersions: func(ctx context.Context, pkgAddr regaddr.ModulePackage) (versions.List, error) {
+		modulePackageVersions: func(ctx context.Context, pkgAddr regaddr.ModulePackage) (ModulePackageVersionsResponse, error) {
+			var ret ModulePackageVersionsResponse
 			for _, pkg := range registryPkgs {
 				if pkg.pkgAddr != pkgAddr {
 					continue
 				}
-				ret := make(versions.List, len(pkg.versions))
+				ret.Versions = make([]ModulePackageInfo, len(pkg.versions))
 				for version := range pkg.versions {
-					ret = append(ret, version)
+					ret.Versions = append(ret.Versions, ModulePackageInfo{
+						Version:     version,
+						Deprecation: registryDeprecations[pkg.pkgAddr.Namespace][version],
+					})
 				}
 				return ret, nil
 			}
-			return nil, fmt.Errorf("no fake registry package matches %s", pkgAddr)
+			return ret, fmt.Errorf("no fake registry package matches %s", pkgAddr)
 		},
-		modulePackageSourceAddr: func(ctx context.Context, pkgAddr regaddr.ModulePackage, version versions.Version) (sourceaddrs.RemoteSource, error) {
+		modulePackageSourceAddr: func(ctx context.Context, pkgAddr regaddr.ModulePackage, version versions.Version) (ModulePackageSourceAddrResponse, error) {
+			var ret ModulePackageSourceAddrResponse
 			for _, pkg := range registryPkgs {
 				if pkg.pkgAddr != pkgAddr {
 					continue
 				}
 				sourceAddr, ok := pkg.versions[version]
 				if !ok {
-					return sourceaddrs.RemoteSource{}, fmt.Errorf("no fake registry package matches %s %s", pkgAddr, version)
+					return ret, fmt.Errorf("no fake registry package matches %s %s", pkgAddr, version)
 				}
-				return sourceAddr, nil
+				ret.SourceAddr = sourceAddr
+				return ret, nil
 			}
-			return sourceaddrs.RemoteSource{}, fmt.Errorf("no fake registry package matches %s", pkgAddr)
+			return ret, fmt.Errorf("no fake registry package matches %s", pkgAddr)
 		},
 	}
 
@@ -716,22 +837,22 @@ func (t *testBuildTracer) appendLogf(f string, v ...interface{}) {
 	t.log = append(t.log, fmt.Sprintf(f, v...))
 }
 
-type packageFetcherFunc func(ctx context.Context, sourceType string, url *url.URL, targetDir string) (*PackageMeta, error)
+type packageFetcherFunc func(ctx context.Context, sourceType string, url *url.URL, targetDir string) (FetchSourcePackageResponse, error)
 
-func (f packageFetcherFunc) FetchSourcePackage(ctx context.Context, sourceType string, url *url.URL, targetDir string) (*PackageMeta, error) {
+func (f packageFetcherFunc) FetchSourcePackage(ctx context.Context, sourceType string, url *url.URL, targetDir string) (FetchSourcePackageResponse, error) {
 	return f(ctx, sourceType, url, targetDir)
 }
 
 type registryClientFuncs struct {
-	modulePackageVersions   func(ctx context.Context, pkgAddr regaddr.ModulePackage) (versions.List, error)
-	modulePackageSourceAddr func(ctx context.Context, pkgAddr regaddr.ModulePackage, version versions.Version) (sourceaddrs.RemoteSource, error)
+	modulePackageVersions   func(ctx context.Context, pkgAddr regaddr.ModulePackage) (ModulePackageVersionsResponse, error)
+	modulePackageSourceAddr func(ctx context.Context, pkgAddr regaddr.ModulePackage, version versions.Version) (ModulePackageSourceAddrResponse, error)
 }
 
-func (f registryClientFuncs) ModulePackageVersions(ctx context.Context, pkgAddr regaddr.ModulePackage) (versions.List, error) {
+func (f registryClientFuncs) ModulePackageVersions(ctx context.Context, pkgAddr regaddr.ModulePackage) (ModulePackageVersionsResponse, error) {
 	return f.modulePackageVersions(ctx, pkgAddr)
 }
 
-func (f registryClientFuncs) ModulePackageSourceAddr(ctx context.Context, pkgAddr regaddr.ModulePackage, version versions.Version) (sourceaddrs.RemoteSource, error) {
+func (f registryClientFuncs) ModulePackageSourceAddr(ctx context.Context, pkgAddr regaddr.ModulePackage, version versions.Version) (ModulePackageSourceAddrResponse, error) {
 	return f.modulePackageSourceAddr(ctx, pkgAddr, version)
 }
 
@@ -757,13 +878,13 @@ func (f stubDependencyFinder) FindDependencies(fsys fs.FS, subPath string, deps
 	file, err := fsys.Open(filePath)
 	if err != nil {
 		if errors.Is(err, fs.ErrNotExist) {
-			diags = diags.Append(&internalDiagnostic{
+			diags = append(diags, &internalDiagnostic{
 				severity: DiagError,
 				summary:  "Missing stub dependency file",
 				detail:   fmt.Sprintf("There is no file %q in the package.", filePath),
 			})
 		} else {
-			diags = diags.Append(&internalDiagnostic{
+			diags = append(diags, &internalDiagnostic{
 				severity: DiagError,
 				summary:  "Invalid stub dependency file",
 				detail:   fmt.Sprintf("Cannot open %q in the package: %s.", filePath, err),
@@ -781,7 +902,7 @@ func (f stubDependencyFinder) FindDependencies(fsys fs.FS, subPath string, deps
 		sourceAddrRaw, versionsRaw, hasVersions := strings.Cut(line, " ")
 		sourceAddr, err := sourceaddrs.ParseSource(sourceAddrRaw)
 		if err != nil {
-			diags = diags.Append(&internalDiagnostic{
+			diags = append(diags, &internalDiagnostic{
 				severity: DiagError,
 				summary:  "Invalid source address in stub dependency file",
 				detail:   fmt.Sprintf("Cannot use %q as a source address: %s.", sourceAddrRaw, err),
@@ -789,7 +910,7 @@ func (f stubDependencyFinder) FindDependencies(fsys fs.FS, subPath string, deps
 			continue
 		}
 		if hasVersions && !sourceAddr.SupportsVersionConstraints() {
-			diags = diags.Append(&internalDiagnostic{
+			diags = append(diags, &internalDiagnostic{
 				severity: DiagError,
 				summary:  "Invalid source address in stub dependency file",
 				detail:   fmt.Sprintf("Cannot specify a version constraint string for %s.", sourceAddr),
@@ -800,7 +921,7 @@ func (f stubDependencyFinder) FindDependencies(fsys fs.FS, subPath string, deps
 		if hasVersions {
 			cnsts, err := constraints.ParseRubyStyleMulti(versionsRaw)
 			if err != nil {
-				diags = diags.Append(&internalDiagnostic{
+				diags = append(diags, &internalDiagnostic{
 					severity: DiagError,
 					summary:  "Invalid version constraints in stub dependency file",
 					detail:   fmt.Sprintf("Cannot use %q as version constraints for %s: %s.", versionsRaw, sourceAddrRaw, err),
@@ -827,7 +948,7 @@ func (f stubDependencyFinder) FindDependencies(fsys fs.FS, subPath string, deps
 		case sourceaddrs.LocalSource:
 			deps.AddLocalSource(sourceAddr, depFinder)
 		default:
-			diags = diags.Append(&internalDiagnostic{
+			diags = append(diags, &internalDiagnostic{
 				severity: DiagError,
 				summary:  "Unsupported source address type",
 				detail:   fmt.Sprintf("stubDependencyFinder doesn't support %T addresses", sourceAddr),
@@ -836,7 +957,7 @@ func (f stubDependencyFinder) FindDependencies(fsys fs.FS, subPath string, deps
 		}
 	}
 	if err := sc.Err(); err != nil {
-		diags = diags.Append(&internalDiagnostic{
+		diags = append(diags, &internalDiagnostic{
 			severity: DiagError,
 			summary:  "Invalid stub dependency file",
 			detail:   fmt.Sprintf("Failed to read %s in the package: %s.", filePath, err),
